#include "reporting_thread.h"
#include "shared_mem_buffer.h"
#include "hwapi.h"

#include <QDateTime>
#include <QDebug>
#include <QApplication>
#include <QCoreApplication>

#include <cmath>
#include <algorithm>

ReportingThread::ReportingThread(hwapi *hw)
  : m_hw(hw)
  , m_fileToDownload(m_hw->dcDownloadFileName()) {
  // , m_f([](QString const&){}) {
}

ReportingThread::~ReportingThread() {
}

// download thread running in ca-slave sends reports of download process to
// each component which has connects for the corresponding signals.
void ReportingThread::run() {

    qCritical()
        << QDateTime::currentDateTime().time().toString(Qt::ISODate)
        << "START REPORT THREAD"
        << "(PART OF APPLICATION" << QCoreApplication::applicationName() << ")";

    static QString report("");

    int cnt = 5;
    while (!m_hw->dcDownloadGetRunning()) {
        if (--cnt > 0) {
            report =  QString("%1 waiting for download to start %2")
                            .arg(QDateTime::currentDateTime().time().toString(Qt::ISODate))
                            .arg(cnt);
            qCritical() << "(" << __func__ << ":" << __LINE__ << ") STATUS" << report;
            emit m_hw->hwapi_reportDCDownloadStatus(report);
            QThread::sleep(1);
        } else break;
    }

#if 0
    if (cnt == 0) {
        m_hw->dcDownloadResetRequest();
        status = QString("%1 reset download request")
                        .arg(QDateTime::currentDateTime().toString(Qt::ISODate));
        qCritical() << __LINE__ << "STATUS" << status;
        emit m_hw->hwapi_reportDCDownloadStatus(status);

        cnt = 5;
        while (!m_hw->dcDownloadRunning()) {
            if (--cnt > 0) {
                QThread::sleep(1);
            } else break;
        }

        if (cnt == 0) {
            status = QString("%1 download request failure")
                            .arg(QDateTime::currentDateTime().toString(Qt::ISODate));
            qCritical() << __LINE__ << "STATUS" << status;
            emit m_hw->hwapi_reportDCDownloadFailure(status);
            return;
        }
    }

    uint16_t const tnr = 1750;
    uint16_t cnr = 0;

    while (cnr <= tnr) {
        QThread::msleep(100);
        QString report("");

        if (cnr > 0) {
            double percent = ((double)cnr / (double)tnr) * 100.0;
            report = QString(": total blocks %1, current block %2 [%3]")
                    .arg(tnr).arg(cnr).arg(percent, 0, 'f', 2);
        } else {
            report = QString(": total blocks %1, current block %2 [0]")
                    .arg(tnr).arg(cnr);
        }
        status = QDateTime::currentDateTime().toString(Qt::ISODate) + report;

        qCritical() << "STATUS" << status;

        emit m_hw->hwapi_reportDCDownloadStatus(status);
        cnr += 1;
    }

    if (tnr == cnr) {
        m_hw->hwapi_reportDCDownloadSuccess(
            QString("SUCCESS DOWNLOADING") + m_fileToDownload);
    } else {
        m_hw->hwapi_reportDCDownloadFailure(
            QString("ERROR DOWNLOADING %1 (total blocks=%2, sent blocks=%3)")
                .arg(m_fileToDownload).arg(tnr).arg(cnr));
    }

#endif

    uint16_t totalBlocks = m_hw->dcDownloadGetTotalBlockNumber();
    cnt = 10;
    while(--cnt > 0 && totalBlocks == 0) {
        totalBlocks = m_hw->dcDownloadGetTotalBlockNumber();
        qCritical() << QDateTime::currentDateTime() << __PRETTY_FUNCTION__
                    << QString("line=%1 TOTAL BLOCKS=%2 (%3)")
                        .arg(__LINE__).arg(totalBlocks).arg(cnt);
        QThread::sleep(1);
    }

    qCritical() << QDateTime::currentDateTime().time().toString(Qt::ISODate) << __PRETTY_FUNCTION__
                << QString("line=%1 TOTAL BLOCKS=%2").arg(__LINE__).arg(totalBlocks);

    if (totalBlocks) {
        QDateTime const start = QDateTime::currentDateTime();
        uint16_t currentBlockNumber = 0;
        uint16_t prevBlockNumber = ~0;
        uint64_t estimatedMinutesLeftMax = ~0ULL;
        uint64_t estimatedSecondsLeftMax = ~0ULL;
        uint64_t estimatedMinutesLeftPrev = 0;
        uint64_t estimatedSecondsLeftPrev = 0;

        while (m_hw->dcDownloadGetRunning()) {
            currentBlockNumber = m_hw->dcDownloadGetCurrentBlockNumber();
            if (prevBlockNumber != currentBlockNumber) {
                double durationSecs = start.secsTo(QDateTime::currentDateTime());

                double const timeAveragePerBlock = (currentBlockNumber > 0) ? (durationSecs / currentBlockNumber) : durationSecs;
                uint64_t estimatedSecondsLeft = lround((timeAveragePerBlock * (totalBlocks - currentBlockNumber)));
                uint64_t estimatedMinutesLeft =
                    ((estimatedSecondsLeft % 60) == 0) ?
                        (estimatedSecondsLeft / 60) :
                            ((estimatedSecondsLeft + 60) / 60);

                estimatedSecondsLeft = (estimatedSecondsLeft % 60);

                if ((estimatedMinutesLeft <= estimatedMinutesLeftMax)
                 || (estimatedSecondsLeft <= estimatedSecondsLeftMax)) {
                    estimatedMinutesLeftMax = estimatedMinutesLeft;
                    estimatedSecondsLeftMax = estimatedSecondsLeft;
                    estimatedMinutesLeftPrev = estimatedMinutesLeft;
                    estimatedSecondsLeftPrev = estimatedSecondsLeft;

                    double percent = ((double)currentBlockNumber / (double)totalBlocks) * 100.0;
                    report = QString(": total blocks %1, current block %2 [%3] (est. time left: %4min %5s)")
                                .arg(totalBlocks)
                                .arg(currentBlockNumber)
                                .arg(percent, 0, 'f', 2)
                                .arg(estimatedMinutesLeft)
                                .arg(estimatedSecondsLeft, 2);

                    qCritical() << QDateTime::currentDateTime().time().toString(Qt::ISODate)
                                << QString("line=%1: RT report").arg(__LINE__) << report;

                    emit m_hw->hwapi_reportDCDownloadStatus(report);

                } else {
                    double percent = ((double)currentBlockNumber / (double)totalBlocks) * 100.0;
                    report = QString(": total blocks %1, current block %2 [%3] (est. time left: %4min %5s)")
                                .arg(totalBlocks)
                                .arg(currentBlockNumber)
                                .arg(percent, 0, 'f', 2)
                                .arg(estimatedMinutesLeftPrev)
                                .arg(estimatedSecondsLeftPrev, 2);

                    qCritical() << QDateTime::currentDateTime().time().toString(Qt::ISODate)
                                << QString("line=%1: RT report").arg(__LINE__) << report;

                    emit m_hw->hwapi_reportDCDownloadStatus(report);
                }

                prevBlockNumber = currentBlockNumber;
            }
            QThread::msleep(100);
        }

        QThread::msleep(100);

        if (totalBlocks == currentBlockNumber) {
            m_hw->hwapi_reportDCDownloadSuccess(
                QString("SUCCESS DOWNLOADING") + m_fileToDownload);
        } else {
            m_hw->hwapi_reportDCDownloadFailure(
                QString("ERROR DOWNLOADING %1 (total blocks=%2, sent blocks=%3)")
                    .arg(m_fileToDownload).arg(totalBlocks).arg(currentBlockNumber));
        }
    }

    qCritical() << QDateTime::currentDateTime().time().toString(Qt::ISODate) << __PRETTY_FUNCTION__
                << QString("line=%1 REPORT THREAD ABOUT TO FINISH").arg(__LINE__);

    cnt = 10;

    bool running = m_hw->dcDownloadGetRunning();
    bool finished = m_hw->dcDownloadGetFinished();

    while (--cnt > 0 && (running && !finished)) {
        qCritical() << QDateTime::currentDateTime() << __PRETTY_FUNCTION__
                    << QString("line=%1 REPORT THREAD: WAIT FOR END OF DOWNLOAD THREAD %2 %3 (%4)")
            .arg(__LINE__).arg(running).arg(finished).arg(cnt);
        QThread::sleep(1);
        running = m_hw->dcDownloadGetRunning();
        finished = m_hw->dcDownloadGetFinished();
    }

    qCritical() << QDateTime::currentDateTime() << __PRETTY_FUNCTION__
                << QString("line=%1 FINISH REPORT THREAD").arg(__LINE__);
}
